<?php
//
// zphp - A Z-machine re-implementation
//
// Copyright (C) 2009-2016 Väinö Helminen
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//

class Z
{
    const HEADER_VERSION = 0x0;
    const HEADER_FLAGS = 0x1;
    const HEADER_RELEASE = 0x2;
    const HEADER_HIGH_MEMORY = 0x4;
    const HEADER_PROGRAM_COUNTER = 0x6;
    const HEADER_DICTIONARY = 0x8;
    const HEADER_STATIC_MEMORY = 0xe;
    const HEADER_OBJECTS = 0xa;
    const HEADER_GLOBALS = 0xc;
    const HEADER_SERIAL = 0x12;
    const HEADER_ABBREVIATIONS = 0x18;

    const OPCODE_0OP = 0;
    const OPCODE_1OP = 1;
    const OPCODE_2OP = 2;
    const OPCODE_VAR = 3;
    const OPCODE_EXT = 4;

    const CONSTANT_LARGE = 0;
    const CONSTANT_SMALL = 1;
    const VARIABLE = 2;
    const OMITTED = 3;

    const OBJ_PARENT = 4;
    const OBJ_SIBLING = 5;
    const OBJ_CHILD = 6;

    const BREAK_NO = 0;
    const BREAK_QUIT = 1;
    const BREAK_INPUT = 2;

    static $OPCODE_TABLE = array(
        self::OPCODE_0OP=>array('0OP', array(
            array('rtrue',0), array('rfalse',0), array('print',0), array('print_ret',0), array('nop',0), array('save',0), array('restore',0),
            array('restart',0), array('ret_popped',0), array('pop',0), array('quit',0), array('newline',0),
            array('show_status',0), array('verify',0))),
        self::OPCODE_1OP=>array('1OP', array(
            array('jz',1), array('get_sibling',1), array('get_child',1), array('get_parent',1),
            array('get_prop_len',1), array('inc',1), array('dec',1), array('print_addr',1), null, array('remove_obj',1),
            array('print_obj',1), array('ret',1), array('jump',1), array('print_paddr',1), array('load',1), array('not',2))),
        self::OPCODE_2OP=>array('2OP', array(
            0, array('je',-2), array('jl',2), array('jg',2), array('dec_chk',2), array('inc_chk',2), array('jin',2), array('test',2),
            array('or',2), array('and',2), array('test_attr',2), array('set_attr',2), array('clear_attr',2), array('store',2),
            array('insert_obj',2), array('loadw',2), array('loadb',2), array('get_prop',2), array('get_prop_addr',2),
            array('get_next_prop',2), array('add',2), array('sub',2), array('mul',2), array('div',2), array('mod',2))),
        self::OPCODE_VAR=>array('VAR', array(
            array('call',-1), array('storew',3), array('storeb',3), array('put_prop',3), array('sread',2), array('print_char',1),
            array('print_num',1), array('random',1), array('push',1), array('pull',1), array('split_window',1),
            array('set_window',1), null, null, null, null, null, null, null, array('output_stream',1),
            array('input_stream',1), array('sound_effect',-1))),
        self::OPCODE_EXT=>array('EXT', array())
    );

    protected static $CHAR_TABLE = array(
        'abcdefghijklmnopqrstuvwxyz',
        'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        "\0\n0123456789.,!?_#'\"/\\-:()"
    );

    protected $filename; // story file
    protected $memory; // z-machine memory
    protected $pc, $fp, $stack; // state (excluding dynamic part of memory)

    protected $abbreviations, $objects, $properties;

    protected $inputline, $status;
    protected $saveRestoreHook;

    /*** public interface ***/

    public function __construct($filename)
    {
        $this->filename = $filename;
        $this->restart();
        $this->resetFlags();
    }

    public function __sleep()
    {
        $this->state = $this->getState();
        return array('filename', 'state');
    }

    public function __wakeup()
    {
        $this->restart();
        $this->setState($this->state);
        unset($this->state);
    }

    public function run(&$status, $inputline=false, $saveRestoreHook=false)
    {
        $this->inputline = $inputline;
        $this->status = false;
        $this->saveRestoreHook = $saveRestoreHook;

        $break = self::BREAK_NO;

        for($counter = 0; !$break and ++$counter < 10000;) {
            $oldpc = $this->pc;

            $this->decodeOpcode($optype, $opcode, $operands);

            $break = $this->executeOpcode($optype, $opcode, $operands);
        }

        if($break == self::BREAK_NO)
            throw new Exception('Infinite loop?');

        if($break == self::BREAK_INPUT)
            // step back last opcode (should be sread)
            $this->pc = $oldpc;

        $status = $this->status;
        return $break;
    }

    /*** private methods ***/

    protected function getState()
    {
        $dynamic = substr($this->memory, 0, $this->dynamicLimit);
        return array($dynamic, $this->pc, $this->fp, $this->stack);
    }

    protected function setState($state)
    {
        list($dynamic, $pc, $fp, $stack) = $state;
        if($this->dynamicLimit != strlen($dynamic) or
           $this->memory[self::HEADER_VERSION] != $dynamic[self::HEADER_VERSION] or
           $this->memory[self::HEADER_RELEASE] != $dynamic[self::HEADER_RELEASE])
            throw new Exception('Invalid game state');
        $this->memory = $dynamic . substr($this->memory, $this->dynamicLimit);
        $this->pc = $pc;
        $this->fp = $fp;
        $this->stack = $stack;
        $this->resetFlags();
    }

    protected function decodeOpcode(&$optype, &$opcode, &$operands)
    {
        $byte = $this->getBytePC();

        if($byte < 0x80) {
            // long opcode, two operands
            $opcode = $byte & 0x1f;
            $optype = self::OPCODE_2OP;
            $operands = array(
                $this->getOperand($byte & 0x40 ? self::VARIABLE : self::CONSTANT_SMALL),
                $this->getOperand($byte & 0x20 ? self::VARIABLE : self::CONSTANT_SMALL)
            );

        } else if(($byte & 0x40) == 0) {
            // short opcode, zero or one operand
            $opcode = $byte & 0xf;
            $operand = $this->getOperand(($byte >> 4) & 3);
            if(false === $operand) {
                $optype = self::OPCODE_0OP;
                $operands = array();
            } else {
                $optype = self::OPCODE_1OP;
                $operands = array($operand);
            }

        } else {
            // variable opcode, variable operands
            $opcode = $byte & 0x1f;
            $optype = $byte < 0xe0 ? self::OPCODE_2OP : self::OPCODE_VAR;
            $types = $this->getBytePC(); // operand types
            $operands = array();
            for($bits = 6; $bits >= 0; $bits -= 2) {
                $op = $this->getOperand(($types >> $bits) & 3);
                if(false === $op) break;
                $operands[] = $op;
            }
        }
    }

    protected function getOperand($type)
    {
        switch($type) {
        case self::CONSTANT_LARGE:
            return $this->getWordPC();

        case self::CONSTANT_SMALL:
            return $this->getBytePC();

        case self::VARIABLE:
            return $this->getVariable($this->getBytePC());

        case self::OMITTED:
        default:
            return false;
        }
    }

    protected function executeOpcode($optype, $opcode, $operands)
    {
        @list($func, $nops) = @self::$OPCODE_TABLE[$optype][1][$opcode];
        if(!$func) return self::BREAK_QUIT;
        return call_user_func_array(array($this, 'z_'.$func), $operands);
    }

    protected function restart()
    {
        $this->memory = file_get_contents($this->filename);
        if(!$this->memory)
            throw new Exception('Failed to read the story file');

        if($this->getByte(self::HEADER_VERSION) != 3)
            throw new Exception('I can only play Z-code version 3 stories');

        // set initial state
        $this->pc = $this->getWord(self::HEADER_PROGRAM_COUNTER);;
        $this->fp = -1;
        $this->stack = array();
        $this->status = false;

        // precalc some stuff
        $this->decodeAbbreviations();
        $addr = $this->getWord(self::HEADER_OBJECTS);
        $this->objects = $addr + 53;
        $this->properties = $addr - 2;
        $this->dynamicLimit = $this->getWord(self::HEADER_STATIC_MEMORY);
    }

    protected function resetFlags()
    {
        // bit 4 = status line is available => 0
        // bit 5 = screen splitting is not available => 0
        // bit 6 = variable with font is not the default (even if it is) => 0
        $this->memory[self::HEADER_FLAGS] = chr($this->getByte(self::HEADER_FLAGS) & 0x8f);
    }

    protected function getByte($addr)
    {
        return ord($this->memory[$addr]);
    }

    protected function setByte($addr, $byte)
    {
        $this->memory[$addr] = chr($byte&0xff);
    }

    protected function getWord($addr)
    {
        return (ord($this->memory[$addr])<<8) | ord($this->memory[$addr+1]);
    }

    protected function setWord($addr, $word)
    {
        $this->memory[$addr] = chr(($word>>8)&0xff);
        $this->memory[$addr+1] = chr(($word)&0xff);
    }

    protected function getBytePC()
    {
        return $this->getByte($this->pc++);
    }

    protected function getWordPC()
    {
        $addr = $this->pc;
        $this->pc += 2;
        return $this->getWord($addr);
    }

    protected function getVariable($var)
    {
        if($var == 0)
            // 0 -> pop a value from stack
            return $this->popStack();
        else if($var < 16)
            // 1-15 -> local variable
            return $this->stack[$this->fp + $var];
        else
            // 16-255 -> global variable
            return $this->getWord($this->getWord(self::HEADER_GLOBALS) + 2 * ($var - 16));
    }

    protected function setVariable($var, $value)
    {
        if($var == 0)
            // 0 -> push the value to the stack
            $this->pushStack($value);
        else if($var < 16)
            // 1-15 -> local variable
            $this->stack[$this->fp + $var] = $value;
        else
            // 16-255 -> global variable
            $this->setWord($this->getWord(self::HEADER_GLOBALS) + 2 * ($var - 16), $value);
    }

    protected function addToVariable($var, $value)
    {
        $value = ($this->getVariable($var) + $value) & 0xffff;
        $this->setVariable($var, $value);
        return $value;
    }

    protected function getObjectAddr($obj)
    {
        return $this->objects + 9 * $obj;
    }

    protected function getPropertyList($obj)
    {
        $addr = $this->getObjectAddr($obj);
        $addr = $this->getWord($addr+7); // pointer to properties
        return $addr + 1 + 2 * $this->getByte($addr); // skip description text
    }

    protected function getPropertySize($addr)
    {
        if(!$addr) return 0;
        $size = $this->getByte($addr);
        return ($size & 0x1f) ? ($size>>5) + 1 : 0;
    }

    protected function getNextProperty($addr)
    {
        $size = $this->getPropertySize($addr);
        return $size ? $addr + 1 + $size : 0;
    }

    protected function getPropertyAddr($obj, $prop)
    {
        $addr = $this->getPropertyList($obj);
        while($addr) {
            $byte = $this->getByte($addr) & 0x1f;
            if($byte <= $prop) {
                if($byte == $prop) return $addr;
            }
            $addr = $this->getNextProperty($addr);
        }
        return 0;
    }

    protected function getDefaultProperty($prop)
    {
        return $this->getWord($this->properties + 2 * $prop);
    }

    protected function getAttributeAddr($obj, $attr)
    {
        return $this->getObjectAddr($obj) + ($attr >> 3);
    }

    protected function pushStack($word)
    {
        array_push($this->stack, $word & 0xffff);
    }

    protected function popStack()
    {
        return array_pop($this->stack);
    }

    protected function signExtend($word)
    {
        if($word & 0x8000)
            return ~((~$word) & 0xffff);
        return $word;
    }

    protected function store($word)
    {
        $this->setVariable($this->getBytePC(), $word);
    }

    protected function makeCall($routine, $args, $retvar=false)
    {
        if($routine == 0) {
            // calling address 0 just returns false
            if(false !== $retvar)
                $this->setVariable($retvar, 0);
            return;
        }
        // remember variable for return value
        $this->stack[] = $retvar;

        // remember old and set new program counter
        $this->stack[] = $this->pc;
        $this->pc = $routine;

        // remember old and set new frame pointer
        $this->stack[] = $this->fp;
        $this->fp = count($this->stack)-1;

        // initialize routine's local variables
        $varnum = $this->getBytePC();
        for($i = 1; $i <= $varnum; ++$i) {
            // get default value for variable
            $value = $this->getWordPC();
            // override value with an argument if given
            if($args) $value = array_shift($args);
            // put variable to the stack where we can find it when needed
            $this->stack[] = $value;
        }

    }

    protected function returnCall($value)
    {
        $fp = $this->fp;
        $this->fp = $this->stack[$fp];
        $this->pc = $this->stack[$fp-1];
        $retvar = $this->stack[$fp-2];
        $this->stack = array_slice($this->stack, 0, $fp-2);
        // set return value if variable was given
        if(false !== $retvar)
            $this->setVariable($retvar, $value);
    }

    protected function branch($flag)
    {
        $offset = $this->getBytePC();

        if(($offset & 0x80) == 0) {
            $flag = !$flag;
        }

        if($offset & 0x40) {
            $offset &= 0x3f;
        } else {
            $offset = ($offset << 8) | $this->getBytePC();
            if($offset & 0x2000) {
                $offset |= 0xc000;
            } else {
                $offset &= 0x3fff;
            }
        }
        if($flag) {
            if($offset == 0 or $offset == 1) {
                $this->returnCall($offset);
            } else {
                $this->z_jump($offset);
            }
        }
    }

    protected function decodeZChar($chr, &$zshift, &$zabbr)
    {
        $shift = $zshift;  $zshift = 0;
        $abbr = $zabbr;  $zabbr = 0;

        if($chr >= 32) {
            // plain ASCII
            return chr($chr);
        }

        if($shift < 0) {
            // ASCII character
            if($shift == -256) {
                // first bits
                $zshift = -32 * $chr;
                return false;
            }
            // second bits
            return chr((-$shift) | $chr);
        }

        if($abbr) {
            // abbreviation
            return $this->abbreviations[$abbr][$chr];
        }

        if($chr < 6) {
            if($chr == 0) {
                // space
                return ' ';
            }
            if($chr < 4) {
                // set abbreviation
                $zabbr = $chr;
                return false;
            }
            // set shift
            $zshift = $chr-3;
            return false;
        }
        if($chr == 6 and $shift == 2) {
            // set ASCII character
            $zshift = -256;
            return false;
        }
        // look up character from the tables
        return static::$CHAR_TABLE[$shift][$chr-6];
    }


    protected function decodeZString(&$addr)
    {
        $zshift = $zabbr = 0;
        $str = array();
        do {
            $word = $this->getWord($addr);
            $addr += 2;
            for($i = 10; $i>=0; $i -= 5) {
                $chr = $this->decodeZChar(($word >> $i)&0x1f, $zshift, $zabbr);
                if(false !== $chr) $str[] = $chr;
            }
        } while(($word & 0x8000) == 0);
        return implode('', $str);
    }


    protected function encodeZString($str, $cutWords)
    {
        // encode to Z characters (upto 3*words characters)
        $z = array();
        $zchars = 3*$cutWords;
        foreach(str_split($str) as $chr) {
            if($chr == ' ') {
                $z[] = 0;
            } else if(false !== ($i = strpos(static::$CHAR_TABLE[0], $chr))) {
                $z[] = $i+6;
            } else if(($i = strpos(static::$CHAR_TABLE[2], $chr)) and $i > 0) {
                $z[] = 5;
                $z[] = $i+6;
            } else if(false !== ($i = strpos(static::$CHAR_TABLE[1], $chr))) {
                $z[] = 4;
                $z[] = $i+6;
            } else {
                $i = ord($chr);
                $z[] = 5;
                $z[] = 6;
                $z[] = $i>>5;
                $z[] = $i&0x1f;
            }
            if(count($z) >= $zchars) break;
        }

        // add padding
        while(count($z) < $zchars) $z[] = 5;

        // combine Z-chars to 16-bit words
        $words = array();
        for($i = 0; $i < $zchars; $i += 3) {
            $words[] = ($z[$i] << 10) | ($z[$i+1] << 5) | ($z[$i+2]);
        }
        // mark the last word
        $words[$cutWords-1] |= 0x8000;

        // get a byte string
        $str = '';
        foreach($words as $word) {
            $str .= chr(($word >> 8) & 0xff) . chr($word & 0xff);
        }
        return $str;
    }

    public function decodeAbbreviations()
    {
        $this->abbreviations = array(1=>array(), 2=>array(), 3=>array());
        $table = $this->getWord(self::HEADER_ABBREVIATIONS);
        for($i = 0; $i < 96; ++$i) {
            $addr = 2 * $this->getWord($table);
            $this->abbreviations[($i >> 5) + 1][] = $this->decodeZString($addr);
            $table += 2;
        }
    }

    protected function getSeparators($addr=0)
    {
        if(!$addr) $addr = $this->getWord(self::HEADER_DICTIONARY);
        return substr($this->memory, $addr+1, $this->getByte($addr));
    }

    protected function getDictionary($addr=0)
    {
        if(!$addr) $addr = $this->getWord(self::HEADER_DICTIONARY);

        // skip separator list
        $addr += $this->getByte($addr) + 1;

        $length = $this->getByte($addr);
        $entries = $this->getWord($addr+1);
        $addr += 3;

        $dictionary = array();
        for($i = 0; $i < $entries; ++$i) {
            $dictionary[substr($this->memory, $addr, 4)] = $addr;
            $addr += $length;
        }
        return $dictionary;
    }

    protected function inputLine($separators)
    {
        $input = $this->inputline;
        $this->inputline = false;

        $chars = $separators.static::$CHAR_TABLE[0].substr(static::$CHAR_TABLE[2],1);
        $chars = str_replace(array('\\', "\n", '/', ']', '[', ':', '^'), '', $chars);
        if(strstr($chars, '-')) $chars = '-'.str_replace('-', '', $chars);
        $chars = '/[^'.$chars.']+/';

        return preg_replace($chars, ' ', strtolower($input));
    }

    /*** OPCODES ***/

    protected function z_rtrue()
    {
        $this->returnCall(1);
    }

    protected function z_rfalse()
    {
        $this->returnCall(0);
    }

    protected function z_print()
    {
        echo $this->decodeZString($this->pc);
    }

    protected function z_print_ret()
    {
        echo $this->decodeZString($this->pc) . "\n";
        $this->returnCall(1);
    }

    protected function z_nop()
    {
        // nop
    }

    protected function z_save()
    {
        if($this->saveRestoreHook and
           call_user_func($this->saveRestoreHook, $this->getState()))
            $this->branch(true);
        else
            $this->branch(false);
    }

    protected function z_restore()
    {
        if($this->saveRestoreHook) {
            $state = call_user_func($this->saveRestoreHook, false);
            if($state) {
                $this->setState($state);
                $this->branch(true);
                return;
            }
        }
        $this->branch(false);
    }

    protected function z_restart()
    {
        $this->restart();
    }

    protected function z_ret_popped()
    {
        $this->returnCall($this->popStack());
    }

    protected function z_pop()
    {
        $this->popStack();
    }

    protected function z_quit()
    {
        return self::BREAK_QUIT;
    }

    protected function z_newline()
    {
        echo "\n";
    }

    protected function z_show_status()
    {
        $this->status = array();
        $addr = $this->getWord($this->getObjectAddr($this->getVariable(16)) + 7) + 1; // get properties address and skip text size byte
        $this->status[] = $this->decodeZString($addr);
        $a = $this->signExtend($this->getVariable(17));
        $b = $this->signExtend($this->getVariable(18));
        if($this->getByte(self::HEADER_FLAGS) & 2) {
            $this->status[] = sprintf('%2d:%02d', $a, $b);
        } else {
            $this->status[] = $a;
            $this->status[] = $b;
        }
    }

    protected function z_verify()
    {
        $this->branch(true); // always succeed
    }

    protected function z_jz($value)
    {
        $this->branch(!$value);
    }

    protected function z_get_sibling($obj)
    {
        $sibling = $this->getByte($this->getObjectAddr($obj) + self::OBJ_SIBLING);
        $this->store($sibling);
        $this->branch($sibling);
    }

    protected function z_get_child($obj)
    {
        $child = $this->getByte($this->getObjectAddr($obj) + self::OBJ_CHILD);
        $this->store($child);
        $this->branch($child);
    }

    protected function z_get_parent($obj)
    {
        $parent = $this->getByte($this->getObjectAddr($obj) + self::OBJ_PARENT);
        $this->store($parent);
    }

    protected function z_get_prop_len($addr)
    {
        $this->store($this->getPropertySize($addr-1));
    }

    protected function z_inc($var)
    {
        $this->addToVariable($var, 1);
    }

    protected function z_dec($var)
    {
        $this->addToVariable($var, -1);
    }

    protected function z_print_addr($addr)
    {
        echo $this->decodeZString($addr);
    }

    protected function z_remove_obj($obj)
    {
        $addr = $this->getObjectAddr($obj);
        $parent = $this->getByte($addr + self::OBJ_PARENT);
        if(!$parent) return;

        $sibling = $this->getByte($addr + self::OBJ_SIBLING);
        $this->setByte($addr + self::OBJ_PARENT, 0); // clear parent
        $this->setByte($addr + self::OBJ_SIBLING, 0); // and sibling

        $addr = $this->getObjectAddr($parent);
        $child = $this->getByte($addr + self::OBJ_CHILD);
        if($obj == $child) {
            // first child. just set parent's child link to object's sibling
            $this->setByte($addr + self::OBJ_CHILD, $sibling);
            return;
        }
        // find object and unlink it from then children list
        do {
            $addr = $this->getObjectAddr($child);
            $child = $this->getByte($addr + self::OBJ_SIBLING);
        } while($child != $obj);
        $this->setByte($addr + self::OBJ_SIBLING, $sibling);
    }

    protected function z_print_obj($obj)
    {
        $addr = $this->getObjectAddr($obj);
        $addr = $this->getWord($addr+7) + 1; // get properties address and skip text size byte
        echo $this->decodeZString($addr);
    }

    protected function z_ret($value)
    {
        $this->returnCall($value);
    }

    protected function z_jump($offset)
    {
        $this->pc += $this->signExtend($offset) - 2;
    }

    protected function z_print_paddr($paddr)
    {
        $this->z_print_addr(2 * $paddr);
    }

    protected function z_load($var)
    {
        $this->store($this->getVariable($var));
    }

    protected function z_not($value)
    {
        $this->store(~$value);
    }

    protected function z_je()
    {
        $a = func_get_arg(0);
        $flag = false;
        $args = func_num_args();
        for($i = 1; $i < $args; ++$i) {
            $b = func_get_arg($i);
            if($a == $b) {
                $flag = true;
                break;
            }
        }
        $this->branch($flag);
    }

    protected function z_jl($a, $b)
    {
        $a = $this->signExtend($a);
        $b = $this->signExtend($b);
        $this->branch($a < $b);
    }

    protected function z_jg($a, $b)
    {
        $a = $this->signExtend($a);
        $b = $this->signExtend($b);
        $this->branch($a > $b);
    }

    protected function z_dec_chk($a, $b)
    {
        $a = $this->addToVariable($a, -1);
        $a = $this->signExtend($a);
        $b = $this->signExtend($b);
        $this->branch($a < $b);
    }

    protected function z_inc_chk($a, $b)
    {
        $a = $this->addToVariable($a, 1);
        $a = $this->signExtend($a);
        $b = $this->signExtend($b);
        $this->branch($a > $b);
    }

    protected function z_jin($obj, $parent)
    {
        $this->branch($this->getByte($this->getObjectAddr($obj) + self::OBJ_PARENT) == $parent);
    }

    protected function z_test($bitmap, $flags)
    {
        $this->branch($bitmap & $flags);
    }

    protected function z_or($a, $b)
    {
        $this->store($a | $b);
    }

    protected function z_and($a, $b)
    {
        $this->store($a & $b);
    }

    protected function z_test_attr($obj, $attr)
    {
        if($obj) {
            $addr = $this->getAttributeAddr($obj, $attr);
            $flag = $this->getByte($addr) & (1 << (7 - ($attr & 7)));
        } else {
            $flag = false;
        }
        $this->branch($flag);
    }

    protected function z_set_attr($obj, $attr)
    {
        $addr = $this->getAttributeAddr($obj, $attr);
        $this->setByte($addr, $this->getByte($addr) | (1 << (7 - ($attr & 7))));
    }

    protected function z_clear_attr($obj, $attr)
    {
        $addr = $this->getAttributeAddr($obj, $attr);
        $this->setByte($addr, $this->getByte($addr) & (~(1 << (7 - ($attr & 7)))));
    }

    protected function z_store($var, $value)
    {
        $this->setVariable($var, $value);
    }

    protected function z_insert_obj($obj, $parent)
    {
        // remove object from where ever it is
        $this->z_remove_obj($obj);

        // set object as parents first child
        $addr = $this->getObjectAddr($parent);
        $sibling = $this->getByte($addr + self::OBJ_CHILD);
        $this->setByte($addr + self::OBJ_CHILD, $obj);

        // set objects parent and sibling
        $addr = $this->getObjectAddr($obj);
        $this->setByte($addr + self::OBJ_PARENT, $parent);
        if($sibling) $this->setByte($addr + self::OBJ_SIBLING, $sibling);
    }

    protected function z_loadw($array, $index)
    {
        $this->store($this->getWord($array + 2 * $index));
    }

    protected function z_loadb($array, $index)
    {
        $this->store($this->getByte($array + $index));
    }

    protected function z_get_prop($obj, $prop)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        switch($this->getPropertySize($addr)) {
        case 1:
            $value = $this->getByte($addr+1);
            break;
        case 2:
            $value = $this->getWord($addr+1);
            break;
        default:
            $value = $this->getDefaultProperty($prop);
            break;
        }
        $this->store($value);
    }

    protected function z_get_prop_addr($obj, $prop)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        $this->store($addr ? $addr + 1 : 0);
    }

    protected function z_get_next_prop($obj, $prop)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        if($prop) $addr = $this->getNextProperty($addr);
        $this->store($this->getByte($addr) & 0x1f);
    }

    protected function z_add($a, $b)
    {
        $this->store($a + $b);
    }

    protected function z_sub($a, $b)
    {
        $this->store($a - $b);
    }

    protected function z_mul($a, $b)
    {
        $this->store($a * $b);
    }

    protected function z_div($a, $b)
    {
        $this->store($a / $b);
    }

    protected function z_mod($a, $b)
    {
        $this->store($a % $b);
    }

    protected function z_call()
    {
        $args = func_get_args();
        $this->makeCall(2 * array_shift($args), $args, $this->getBytePC());
    }

    protected function z_storew($array, $index, $value)
    {
        $this->setWord($array + 2 * $index, $value);
    }

    protected function z_storeb($array, $index, $value)
    {
        $this->setByte($array + $index, $value);
    }

    protected function z_put_prop($obj, $prop, $value)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        switch($this->getPropertySize($addr)) {
        case 1:
            $this->setByte($addr + 1, $value);
            break;
        case 2:
            $this->setWord($addr + 1, $value);
            break;
        }
    }

    protected function z_sread($textbuf, $parsebuf)
    {
        if(!$this->inputline) {
            // break out to get input from user
            $this->z_show_status();
            return self::BREAK_INPUT;
        }

        $separators = $this->getSeparators();
        $input = $this->inputLine($separators);
        $dictionary = $this->getDictionary();

        $maxchars = $this->getByte($textbuf);
        $maxwords = $this->getByte($parsebuf);
        if($maxwords > $maxchars/2) $maxwords = ($maxwords - 2) >> 2;

        $words = $chars = 0;

        foreach(preg_split('/\s+|(['.$separators.'])/', $input, $maxwords, PREG_SPLIT_DELIM_CAPTURE) as $word) {
            if(!$word) continue;

            // check if the word is in the dictionary
            $zword = $this->encodeZString($word, 2);
            $addr = intval(@$dictionary[$zword]);

            $this->setWord($parsebuf + 4 * $words + 2, $addr); // byte address of dictionary word
            $this->setByte($parsebuf + 4 * $words + 4, strlen($word)); // word length
            $this->setByte($parsebuf + 4 * $words + 5, 1 + $chars); // offset in text buffer
            $words += 1;

            // write the word to the text buffer
            foreach(str_split($word) as $chr) {
                if($chars < $maxchars) {
                    $this->memory[$textbuf + 1 + $chars] = $chr;
                    $chars += 1;
                }
            }
            // separate words with a space
            if($chars < $maxchars) {
                $this->memory[$textbuf + 1 + $chars] = ' ';
                $chars += 1;
            }
        }
        $this->memory[$textbuf + 1 + $chars] = "\0"; // null terminate text buffer
        $this->setByte($parsebuf + 1, $words); // total number of words written
    }

    protected function z_print_char($chr)
    {
        echo chr($chr);
    }

    protected function z_print_num($number)
    {
        echo $this->signExtend($number);
    }

    protected function z_random($range)
    {
        if(($range == 0) or ($range & 0x8000)) {
            // cannot set the seed
            $value = 0;
        } else {
            $value = mt_rand(1, $range); // inclusive 1..range
        }
        $this->store($value);
    }

    protected function z_push($a)
    {
        $this->pushStack($a);
    }

    protected function z_pull($a)
    {
        $this->setVariable($a, $this->popStack());
    }

    protected function z_split_window($window)
    {
        // do not have windows...
    }

    protected function z_set_window($window)
    {
        // do not have windows...
    }

    protected function z_output_stream($stream)
    {
        // currently just ignored
    }

    protected function z_input_stream($stream)
    {
        // currently just ignored
    }

    protected function z_sound_effect()
    {
        // do not have sound...
    }
} // Z
