<?php
//
// Debugging classes for zphp
//
// Copyright (C) 2009-2016 Väinö Helminen
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//

require_once(__DIR__.'/z.php');

class ZCheck extends Z
{
    protected function executeOpcode($optype, $opcode, $operands)
    {
        // check that opcode is valid
        $optable = @self::$OPCODE_TABLE[$optype];
        @list($func, $nops) = @$optable[1][$opcode];
        if(!$func)
            throw new Exception(sprintf('Unknown opcode %s %x', $optable ? $optable[0] : '???', $opcode));

        // check that operand count is correct for this opcode
        if($nops < 0) {
            if(count($operands) < -$nops)
                throw new Exception(sprintf('%s: need at least %d operands, %d given', $func, $nops, count($operands)));
        } else if(count($operands) != $nops) {
            throw new Exception(sprintf('%s: need %d operands, %d given', $func, $nops, count($operands)));
        }

        $break = parent::executeOpcode($optype, $opcode, $operands);
        if($break < self::BREAK_NO or $break > self::BREAK_INPUT)
            throw new Exception('Invalid break code from opcode function');
        return $break;
    }

    protected function getByte($addr)
    {
        if($addr < 0 or $addr >= strlen($this->memory))
            throw new Exception('Invalid memory read');
        return parent::getByte($addr);
    }

    protected function setByte($addr, $byte)
    {
        if($addr < 0 or $addr >= $this->dynamicLimit)
            throw new Exception('Invalid memory write');
        parent::setByte($addr, $byte);
    }

    protected function getWord($addr)
    {
        if($addr < 0 or $addr >= strlen($this->memory))
            throw new Exception('Invalid memory read');
        return parent::getWord($addr);
    }

    protected function setWord($addr, $word)
    {
        if($addr < 0 or $addr >= $this->dynamicLimit - 2)
            throw new Exception('Invalid memory write');
        parent::setWord($addr, $word);
    }

    protected function getVariable($var)
    {
        if($var < 0 or $var > 255)
            throw new Exception('Invalid variable '.$var);
        if($var > 0 and $var < 16 and count($this->stack) <= $this->fp + $var)
            throw new Exception('No local variable '.$var);
        return parent::getVariable($var);
    }

    protected function setVariable($var, $value)
    {
        if($var < 0 or $var > 255)
            throw new Exception('Invalid variable '.$var);
        if($var > 0 and $var < 16 and count($this->stack) <= $this->fp + $var)
            throw new Exception('No local variable '.$var);
        parent::setVariable($var, $value);
    }

    protected function getObjectAddr($obj)
    {
        if($obj < 1 or $obj > 255)
            throw new Exception('Invalid object '.$obj);
        return parent::getObjectAddr($obj);
    }

    protected function getDefaultProperty($prop)
    {
        if($prop < 1 or $prop > 31)
            throw new Exception('Invalid property #'.$prop);
        return parent::getDefaultProperty($prop);
    }

    protected function getAttributeAddr($obj, $attr)
    {
        if($attr < 0 or $attr > 31)
            throw new Exception('Invalid attribute #'.$attr);
        return parent::getAttributeAddr($obj, $attr);
    }

    protected function popStack()
    {
        if(!$this->stack or count($this->stack) <= $this->fp + 1)
            throw new Exception('Pop from an empty stack');
        return parent::popStack();
    }

    protected function returnCall($value)
    {
        if($this->fp < 0)
            throw new Exception('Return from "main" routine');
        if(count($this->stack) <= $this->fp)
            throw new Exception('Stack has been corrupted');
        parent::returnCall($value);
    }

    protected function z_get_prop($obj, $prop)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        if($this->getPropertySize($addr) > 2)
            throw new Exception('get_prop: illegal property size');
        parent::z_get_prop($obj, $prop);
    }

    protected function z_get_next_prop($obj, $prop)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        if(!$addr) throw new Exception('get_next_prop: invalid property');
        parent::z_get_next_prop($obj, $prop);
    }

    protected function z_div($a, $b)
    {
        if($b == 0) throw new Excepetion('div: division by zero');
        parent::z_div($a, $b);
    }

    protected function z_mod($a, $b)
    {
        if($b == 0) throw new Excepetion('mod: division by zero');
        parent::z_mod($a % $b);
    }

    protected function z_call()
    {
        $args = func_get_args();
        $routine = 2 * $args[0];
        if($routine != 0 and ($routine < 0x40 or $routine >= strlen($this->memory)))
            throw new Exception('call: illegal address');

        $varnum = $this->getByte($routine);
        if($varnum > 15)
            throw new Exception('call: too many local variables '.$varnum);
        #if($varnum < count($args)-1)
        #    throw new Exception('call: too many arguments: '.var_export($args, true));

        call_user_func_array(array($this, 'parent::z_call'), $args);
    }

    protected function z_put_prop($obj, $prop, $value)
    {
        $addr = $this->getPropertyAddr($obj, $prop);
        $size = $this->getPropertySize($addr);
        if(!$size)
            throw new Exception('put_prop: property does not exist');
        if($size > 2)
            throw new Exception('put_prop: illegal property size');
        parent::z_put_prop($obj, $prop, $value);
    }

    protected function z_print_char($chr)
    {
        if($chr < 1 or $chr > 255)
            throw new Exception('print_char: invalid character');
        parent::z_print_char($chr);
    }

} // ZCheck


class ZDebug extends ZCheck
{
    public function __construct($filename)
    {
        parent::__construct($filename);
        $this->stdout = fopen('php://stdout', 'w');
        $this->row = 0;
        $this->skiprow = false;
    }

    public function DEBUG()
    {
        $num = func_num_args();
        if($num == 0) $line == 'DEBUG';
        else if($num == 1) $line = func_get_arg(0);
        else {
            $args = func_get_args();
            $line = call_user_func_array('sprintf', $args);
        }
        fwrite($this->stdout, $line."\n");
    }

    protected function executeOpcode($optype, $opcode, $operands)
    {
        if(!$this->skiprow) {
            $optable = @self::$OPCODE_TABLE[$optype];

            if(1)
                $line = sprintf("PC: %04x %s %02x %-16s ", $this->pc, ($optable?$optable[0]:'---'), $opcode, $optable ? (@$optable[1][$opcode] ? $optable[1][$opcode][0] : '-') : '-');
            else
                $line = sprintf("PC: %04x %s %02x %-16s ", $this->pc, ($optable?$optable[0]:'---'), $opcode, '');

            for($i = 0; $i < 4; ++$i) {
                if(count($operands) > $i)
                    $line .= sprintf("%s%04x", $i?',':' ', $operands[$i]);
                else
                    $line .= "     ";
            }
            $line .= "   #" . (++$this->row);

            $this->DEBUG($line);
        } else {
            $this->skiprow = false;
        }
        $break = parent::executeOpcode($optype, $opcode, $operands);

        if($break == self::BREAK_INPUT) $this->skiprow = true;

        return $break;
    }

    protected function branch($flag)
    {
        $off = $this->getByte($this->pc);
        $this->DEBUG("branch(%d,%02x)", $flag, $off);
        parent::branch($flag);
    }

    protected function store($value)
    {
        $var = $this->getByte($this->pc);
        $this->DEBUG("store(%d,%04x)", $var, $value);
        parent::store($value);
    }

    protected function z_test_attr($obj, $attr)
    {
        if($obj == 0) $this->DEBUG('test_attr: object=0');
        parent::z_test_attr($obj, $attr);
    }
} // ZDebug
